@@ -173,7 +173,7 @@ class Step(Registrable, Generic[T]):
         step_name: Optional[str] = None,
         cache_results: Optional[bool] = None,
         step_format: Optional[Format] = None,
-        step_config: Optional[Dict[str, Any]] = None,
+        step_config: Optional[Union[Dict[str, Any], Params]] = None,
         step_unique_id_override: Optional[str] = None,
         step_resources: Optional[StepResources] = None,
         step_metadata: Optional[Dict[str, Any]] = None,
@@ -249,7 +249,11 @@ class Step(Registrable, Generic[T]):
         self.work_dir_for_run: Optional[
             Path
         ] = None  # This is set only while the run() method runs.
-        self._config = step_config
+        if isinstance(step_config, Params):
+            self._config = step_config.as_dict(quiet=True)
+        else:
+            self._config = step_config
+        assert step_resources is None or isinstance(step_resources, StepResources)
         self.step_resources = step_resources
         self.metadata = deepcopy(self.METADATA)
         if step_metadata:
@@ -541,8 +545,8 @@ class Step(Registrable, Generic[T]):
             return False
 
     def _replace_steps_with_results(self, o: Any, workspace: "Workspace"):
-        if isinstance(o, Step):
-            return o.result(workspace, self)
+        if isinstance(o, (Step, StepIndexer)):
+            return o.result(workspace=workspace, needed_by=self)
         elif isinstance(o, Lazy):
             return Lazy(
                 o._constructor,
@@ -623,6 +627,8 @@ class Step(Registrable, Generic[T]):
             elif isinstance(o, WithUnresolvedSteps):
                 yield from dependencies_internal(o.args)
                 yield from dependencies_internal(o.kwargs)
+            elif isinstance(o, StepIndexer):
+                yield o.step
             elif isinstance(o, str):
                 return  # Confusingly, str is an Iterable of itself, resulting in infinite recursion.
             elif isinstance(o, (dict, Params)):
@@ -696,11 +702,23 @@ class Step(Registrable, Generic[T]):
                 self.name,
             )
 
-    def log_failure(self, exception: BaseException) -> None:
-        log_exception(exception, logger=self.logger)
+    def log_failure(self, exception: Optional[BaseException] = None) -> None:
+        if exception is not None:
+            log_exception(exception, logger=self.logger)
         cli_logger.error('[red]\N{ballot x} Step [bold]"%s"[/] failed[/]', self.name)
 
 
+class StepIndexer:
+    def __init__(self, step: Step, key: Union[str, int]):
+        self.step = step
+        self.key = key
+
+    def result(
+        self, workspace: Optional["Workspace"] = None, needed_by: Optional["Step"] = None
+    ) -> Any:
+        return self.step.result(workspace=workspace, needed_by=needed_by)[self.key]
+
+
 class WithUnresolvedSteps(CustomDetHash):
     """
     This is a helper class for some scenarios where steps depend on other steps.
@@ -794,8 +812,8 @@ class WithUnresolvedSteps(CustomDetHash):
         :return: A new object that's a copy of the original object, with all instances of :class:`.Step` replaced
                  with the results of the step.
         """
-        if isinstance(o, Step):
-            return o.result(workspace)
+        if isinstance(o, (Step, StepIndexer)):
+            return o.result(workspace=workspace)
         elif isinstance(o, Lazy):
             return Lazy(
                 o._constructor,
